/*

    sexpr -- python module for scheme symbolic expressions

    Copyright (C) 1999  Eric S. Tiedemann

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact: Eric S. Tiedemann, est@hyperreal.org

*/

/*

make getattr work
proper __str__ and __repr__ and print
better write and display methods
better return codes in prsexpr
add ints and bools and chars
use PyLong_FromString()
use this in mixd perhaps
use unsafe macros when it's safe :)
thread friendliness in print methods
incorp more xxmodule details
separate eof type (we use the fact that it's the only TOK type to 
  escape to the user at present)
fix 1K limit in symbol, number and (esp) string readin (use py strings?)
use Py_Malloc() and Py_Free() where appropriate
break out sexpr.pair, sexpr.symbol etc into separate types
then we can have sexpr.rational :)
package and post this some time
handle 100% of r5rs lexics

*/

#include <python1.5/Python.h>
#include <python1.5/object.h>

#include <ctype.h>
#include <stdlib.h>
#include <errno.h>

/*#define DEBUG*/

typedef struct
{
  PyObject_HEAD
  enum sexpr_type {SEXPR_NULL, SEXPR_PAIR, SEXPR_SYMBOL, SEXPR_TOK} 
  t;
  union
  {
    struct {PyObject *a, *b;} pval;
    PyObject *sval;
    double fval;
  } v;
} sexpr;

static PyObject *obarray;

static PyObject *SexprError;

static PyTypeObject sexprtype;

#define is_sexpr(o) ((o)->ob_type == &sexprtype)

static sexpr *
mksexpr()
{
  sexpr *self = PyObject_NEW(sexpr, &sexprtype);
  // this is just in case
  self->t = SEXPR_NULL;
  return self;
}

PyObject *nullsexpr;

PyObject *
mknullsexpr()
{
  nullsexpr = (PyObject *) mksexpr();
  if (!nullsexpr)
    return 0;
  ((sexpr *) nullsexpr)->t = SEXPR_NULL;
  return nullsexpr;
}

PyObject *
mktoksexpr()
{
  sexpr *s = mksexpr();
  if (!s)
    return 0;
  s->t = SEXPR_TOK;
  return (PyObject *) s;
}

static PyObject *
newsexpr(PyObject *ignore, PyObject *args)
{
  sexpr *self = mksexpr();

  if (!self)
    return 0;

  /* any init */

  return (PyObject *) self;
}

static void
delsexpr(sexpr *self)
{
#ifdef DEBUG
  fprintf(stderr, "starting delsexpr(%p) refcount = %d\n", 
	  self, ((PyObject *) self)->ob_refcnt);
#endif
  if (self->t == SEXPR_PAIR)
    {
#ifdef DEBUG
      fprintf(stderr, "%p is a pair\n", self);
      fprintf(stderr, "%p is a pair; car has count %d; cdr has count %d\n", 
	      self, self->v.pval.a->ob_refcnt, self->v.pval.b->ob_refcnt);
#endif
      Py_DECREF(self->v.pval.a);
      Py_DECREF(self->v.pval.b);
    }
  else if (self->t == SEXPR_SYMBOL)
    {
#ifdef DEBUG
      fprintf(stderr, "%p is a symbol or string; sval member has count %d\n", 
	      self, self->v.sval->ob_refcnt);
#endif
      Py_DECREF(self->v.sval);
    }
  PyMem_DEL(self);
}

static int
issubtype0(enum sexpr_type t, sexpr *o)
{
  return o->t == t;
}

static int prsexpr(PyObject *self, FILE *fp, int flags);

static void
prtail(sexpr *l, FILE *fp, int flags)
{
#ifdef DEBUG
  fprintf(stderr, "entering prtail(%p)\n", l);
#endif
  if (issubtype0(SEXPR_PAIR, l))
    {
      fprintf(fp, " ");
      prsexpr(l->v.pval.a, fp, flags);
      prtail((sexpr *) l->v.pval.b, fp, flags);
    }
  else if (issubtype0(SEXPR_NULL, l))
    fprintf(fp, ")");
  else
    {
      fprintf(fp, " . ");
      prsexpr((PyObject *) l, fp, flags);
      fprintf(fp, ")"); 
    }
}

static void
prstring(PyObject *p, FILE *fp, int flags)
{
  const char *s = PyString_AsString(p);

  putc('\"', fp);
  while (*s)
    {
      if (*s == '\\' || *s == '\"')
	fprintf(fp, "\\%c", *s);
      else
	putc(*s, fp);

      s++;
    }
  putc('\"', fp);
}

static int
prsexpr(PyObject *o, FILE *fp, int flags)
{
#ifdef DEBUG
  fprintf(stderr, "entering prsexpr(%p)\n", o);
#endif
  if (is_sexpr(o))
    {
      sexpr * self = (sexpr *) o;

#ifdef DEBUG
      fprintf(stderr, "prsexpr: object is sexpr of type %d\n", self->t);
#endif

      switch (self->t)
	{
	case SEXPR_NULL: 
	  fprintf(fp, "()");
	  break;
	case SEXPR_PAIR: 
	  fprintf(fp, "("); 
	  prsexpr(self->v.pval.a, fp, flags);
	  prtail((sexpr *) self->v.pval.b, fp, flags);
	  break;
	case SEXPR_SYMBOL:
#ifdef DEBUG
      fprintf(stderr, "prsexpr: sexpr is symbol\n");
#endif
	  fprintf(fp, "%s", PyString_AsString(self->v.sval));
	  break;
	case SEXPR_TOK:
	  fprintf(fp, "#<weird>");
	  break;
	}

      return 0;
    }
  else if (PyString_Check(o))
    /* do this to scheme escapify python strings */
    {
      prstring(o, fp, flags);
      return 0;
    }
  else
    /* recursive calls may run across other python objects */
    return PyObject_Print(o, fp, flags);
}

static PyObject *
sexpr_pairp(PyObject *self, PyObject *args)
{
  PyObject *v;

  if (args && !PyArg_ParseTuple(args, ""))
    return 0;

  if (((sexpr *) self)->t != SEXPR_PAIR)
    return Py_BuildValue("i", 0);
  else
    return Py_BuildValue("i", 1);
}

static struct PyMethodDef sexpr_methods[] = {
  {"pairp", sexpr_pairp},
  {NULL, NULL}
};

static PyObject *
sexpr_getattr(PyObject *self, char *name)
{
  if (strcmp(name, "car") == 0)
    {
      PyObject *v;

      if (((sexpr *) self)->t != SEXPR_PAIR)
	{
	  PyErr_SetString(PyExc_TypeError, "sexpr isn't a pair");
	  return 0;
	}

      v = ((sexpr *) self)->v.pval.a;
      
      Py_INCREF(v);
      return v;
    }
  if (strcmp(name, "cdr") == 0)
    {
      PyObject *v;

      if (((sexpr *) self)->t != SEXPR_PAIR)
	{
	  PyErr_SetString(PyExc_TypeError, "sexpr isn't a pair");
	  return 0;
	}

      v = ((sexpr *) self)->v.pval.b;
      
      Py_INCREF(v);
      return v;
    }
  if (strcmp(name, "cadr") == 0)
    {
      PyObject *v;

      if (((sexpr *) self)->t != SEXPR_PAIR)
	{
	  PyErr_SetString(PyExc_TypeError, "sexpr isn't a pair");
	  return 0;
	}

      v = ((sexpr *) self)->v.pval.b;

      if (((sexpr *) v)->t != SEXPR_PAIR)
	{
	  PyErr_SetString(PyExc_TypeError, "sexpr's cdr isn't a pair");
	  return 0;
	}

      v = ((sexpr *) v)->v.pval.a;

      Py_INCREF(v);
      return v;
    }
  if (strcmp(name, "caddr") == 0)
    {
      PyObject *v;

      if (((sexpr *) self)->t != SEXPR_PAIR)
	{
	  PyErr_SetString(PyExc_TypeError, "sexpr isn't a pair");
	  return 0;
	}

      v = ((sexpr *) self)->v.pval.b;

      if (((sexpr *) v)->t != SEXPR_PAIR)
	{
	  PyErr_SetString(PyExc_TypeError, "sexpr's cdr isn't a pair");
	  return 0;
	}

      v = ((sexpr *) v)->v.pval.b;

      if (((sexpr *) v)->t != SEXPR_PAIR)
	{
	  PyErr_SetString(PyExc_TypeError, "sexpr's cddr isn't a pair");
	  return 0;
	}

      v = ((sexpr *) v)->v.pval.a;

      Py_INCREF(v);
      return v;
    }
  else
    return Py_FindMethod(sexpr_methods, self, name);
}

static PyObject *
sexpr_str(PyObject *self)
{
  PyObject *v;

  if (((sexpr *) self)->t != SEXPR_SYMBOL)
    {
      v = 
      v = ((sexpr *) self)->v.sval;
      
      Py_INCREF(v);
      return v;
    }

  /*  else
    {
      return Py_BuildValue("s", "<sexpr>");
      }*/
  return 0;
}

static PyTypeObject sexprtype =
{
  PyObject_HEAD_INIT(&PyType_Type)
  0,
  "sexpr",
  sizeof(sexpr),
  0,

  (destructor) delsexpr,
  (printfunc) prsexpr,
  sexpr_getattr,
  0,
  0,
  0,

  0,
  0,
  0,

  0,
  0,
  sexpr_str,
};

static PyObject *
getnull(PyObject *self, PyObject *args)
{
  if (PyTuple_Size(args) != 0)
    return 0;

  Py_INCREF(nullsexpr);
  return (PyObject *) nullsexpr;
}

static PyObject *
sexprp(PyObject *self, PyObject *args)
{
  PyObject *o;

  if (!PyArg_ParseTuple(args, "O", &o))
    return 0;

  if (is_sexpr(o))
    return o;
  else
    {
      Py_INCREF(Py_None);
      return Py_None;
    }
}

static PyObject *
issubtype(enum sexpr_type t, PyObject *args)
{
  PyObject *o;

  if (!PyArg_ParseTuple(args, "O", &o))
    return 0;

  if (is_sexpr(o) && ((sexpr *) o)->t == t)
    {
      Py_INCREF(o);
      return o;
    }
  else
    {
      Py_INCREF(Py_None);
      return Py_None;
    }
}

static PyObject *
pairp(PyObject *self, PyObject *args)
{  return issubtype(SEXPR_PAIR, args);}

static PyObject *
nullp(PyObject *self, PyObject *args)
{  return issubtype(SEXPR_NULL, args);}

static PyObject *
eofobjectp(PyObject *self, PyObject *args)
{  return issubtype(SEXPR_TOK, args);}

static PyObject *
symbolp(PyObject *self, PyObject *args)
{  return issubtype(SEXPR_SYMBOL, args);}

static PyObject *
really_cons(PyObject *a, PyObject *b)
{
  sexpr *self = mksexpr();

  if (!self)
    return 0;

  self->t = SEXPR_PAIR;
  self->v.pval.a = a;
  self->v.pval.b = b;

  return (PyObject *) self;
}

static PyObject *
cons(PyObject *ignore, PyObject *args)
{
  PyObject *a, *b, *v;

  if (!PyArg_ParseTuple(args, "OO", &a, &b))
    return 0;

  if (!is_sexpr(a) || !is_sexpr(b))
    {
      PyErr_SetString(PyExc_TypeError, "argument isn't a sexpr");
      return 0;
    }

  v = really_cons(a, b);
  Py_INCREF(a);
  Py_INCREF(b);
  return v;
}

static PyObject *
really_intern(char *s)
{
  PyObject *x;
  sexpr *self;
  PyObject *s1;

  if ((x = PyDict_GetItemString(obarray, s)))
    {
      Py_INCREF(x);
      return x;
    }

  self = mksexpr();

  if (!self)
    return 0;
  
  if (!(s1 = PyString_FromString(s)))
    {
      PyMem_DEL(self);
      return 0;
    }

  self->t = SEXPR_SYMBOL;
  self->v.sval = s1;

  if (PyDict_SetItemString(obarray, s, (PyObject *) self) == -1)
    {
      Py_DECREF(s1);
      PyMem_DEL(self);
      return 0;
    }

  return (PyObject *) self;
}

static PyObject *
intern(PyObject *ignore, PyObject *args)
{
  char *s;

  if (!PyArg_ParseTuple(args, "s;sexpr.intern", &s))
    return 0;

  return really_intern(s);
}

static PyObject *
car(PyObject *ignore, PyObject *args)
{
  PyObject *o, *v;

  if (!PyArg_ParseTuple(args, "O", &o))
    return 0;

  if (!is_sexpr(o))
    {
      PyErr_SetString(PyExc_TypeError, "argument isn't a sexpr");
      return 0;
    }

  if (((sexpr *) o)->t != SEXPR_PAIR)
    {
      PyErr_SetString(PyExc_TypeError, "sexpr isn't a pair");
      return 0;
    }

  v = ((sexpr *) o)->v.pval.a;

  Py_INCREF(v);
  return v;
}

static PyObject *
cdr(PyObject *ignore, PyObject *args)
{
  PyObject *o, *v;

  if (!PyArg_ParseTuple(args, "O", &o))
    return 0;

  if (!is_sexpr(o))
    {
      PyErr_SetString(PyExc_TypeError, "argument isn't a sexpr");
      return 0;
    }

  if (((sexpr *) o)->t != SEXPR_PAIR)
    {
      PyErr_SetString(PyExc_TypeError, "sexpr isn't a pair");
      return 0;
    }

  v = ((sexpr *) o)->v.pval.b;

  Py_INCREF(v);
  return v;
}

static PyObject *eofobject;
static PyObject *openp;
static PyObject *closep;
static PyObject *dot;

static PyObject *read1(FILE *f);
static PyObject *read0(FILE *f);

static PyObject *
read_list(FILE *f)
{
  PyObject *o = read1(f);

  if (!o)
    return 0;

  if (o == eofobject)
  {
    PyErr_SetString(PyExc_EOFError, "in parsing symbolic expression");
    Py_DECREF(o);
    return 0;
  }
  if (o == closep)
    {
      Py_DECREF(o);
      Py_INCREF(nullsexpr);
      return nullsexpr;
    }
  else if (o == dot)
    {
      Py_DECREF(o);
      o = read0(f);
      if (o == eofobject)
	{
	  PyErr_SetString(PyExc_EOFError, "in parsing symbolic expression");
	  Py_DECREF(o);
	  return 0;
	}
      else
	{
	  PyObject *e;

	  e = read1(f);
	  if (e == closep)
	    {
	      Py_DECREF(e);
	      return o;
	    }
	  else if (e == eofobject)
	    {
	      PyErr_SetString(PyExc_EOFError, 
			      "in parsing symbolic expression");
	      Py_DECREF(e);
	      Py_DECREF(o);
	      return 0;
	    }
	  else
	    {
	      PyErr_SetString(PyExc_SyntaxError, 
			      "in parsing symbolic expression");
	      Py_DECREF(e);
	      Py_DECREF(o);
	      return 0;
	    }
	}
    }
  else
    {
      PyObject *b;
      PyObject *o1;
      b = read_list(f);
      if (b)
	return really_cons(o, b);
      else
	{
	  Py_DECREF(o);
	  return 0;
	}
    }
}

static PyObject *
rdstring(FILE *fp)
{
  char buf[1024];
  int bufi;
  int c;

  bufi = 0;

  c = getc(fp);
  if (c == EOF)
    {
      PyErr_SetString(PyExc_EOFError, "while reading string");
      return 0;
    }

  while (c != '\"' && bufi < sizeof(buf) - 1)
    {
      if (c == '\\')
	{
	  int c = getc(fp);
	  if (c == EOF)
	    {
	      PyErr_SetString(PyExc_EOFError, "while reading string");
	      return 0;
	    }
	  else if (c == 'n')
	    buf[bufi++] = '\n';
	  else if (c == 't')
	    buf[bufi++] = '\t';
	  else
	    buf[bufi++] = c;
	}
      else
	buf[bufi++] = c;

      c = getc(fp);
      if (c == EOF)
	{
	  PyErr_SetString(PyExc_EOFError, "while reading string");
	  return 0;
	}
    }

  buf[bufi] = 0;

  return PyString_FromString(buf); /*mkstring(buf);*/
}

static PyObject *
rdatom(FILE *fp)
{
  int c;
  double d;
  char *end;
  char buf[1024];
  int bufi;

  bufi = 0;

  while (bufi < sizeof(buf) - 1)
    {
      Py_BEGIN_ALLOW_THREADS
      c = getc(fp);
      Py_END_ALLOW_THREADS
      switch (c)
	{
	case EOF: goto looped;
	case '(':
	  if (bufi == 0)
	    {
	      Py_INCREF(openp);
	      return openp;
	    }
	  else
	    {
	      ungetc(c, fp);
	      goto looped;
	    }
	case ')':
	  if (bufi == 0)
	    {
	      Py_INCREF(closep);
	      return closep;
	    }
	  else
	    {
	      ungetc(c, fp);
	      goto looped;
	    }
	case '\"':
	  if (bufi == 0)
	    return rdstring(fp);
	  else
	    {
	      ungetc(c, fp);
	      goto looped;
	    }
	case ';':
	  if (bufi == 0)
	    {
	      Py_BEGIN_ALLOW_THREADS
	      while ((c = getc(fp)) != EOF && c != '\n')
		;			/* consume comment */
	      Py_END_ALLOW_THREADS
	    }
	  else
	    {
	      ungetc(c, fp);
	      goto looped;
	    }
	default:
	  if (isspace(c))
	    {
	      if (bufi > 0)
		goto looped;
	    }
	  else
	    buf[bufi++] = c;
	}
    }

 looped:
  buf[bufi] = 0;

  if (bufi == 0)
    {
      Py_INCREF(eofobject);
      return eofobject;
    }

  if (bufi == 1 && *buf == '.')
    {
      Py_INCREF(dot);
      return dot;
    }
    
  d = strtod(buf, &end);

  if (end == buf + strlen(buf))
    return PyFloat_FromDouble(d);
  else
    return really_intern(buf);
}

static PyObject *
read1(FILE *f)
{
  PyObject *o = rdatom(f);

  if (o == openp)
    {
      Py_DECREF(o);
      return read_list(f);
    }
  else
    return o;
}

static PyObject *
read0(FILE *f)
{
  PyObject *o = read1(f);
  if (!o)
    return 0;
  else if (o == dot)
  {
    PyErr_SetString(PyExc_SyntaxError, "superfluous period");
    Py_DECREF(o);
    return 0;
  }
  else if (o == closep)
  {
    PyErr_SetString(PyExc_SyntaxError, "superfluous close parenthesis");
    Py_DECREF(o);
    return 0;
  }
  else
    return(o);
}

static PyObject *sysmod, *sysmodd;

static PyObject *
writesexpr(PyObject *self, PyObject *args)
{
  PyObject *o;

  if (! PyArg_ParseTuple(args, "O;sexpr.write", &o))
    return 0;

  prsexpr(o, stdout, 0);

  return o;
}

static PyObject *
readsexpr(PyObject *self, PyObject *args)
{
  PyObject *f, *o;

  if (PyTuple_Size(args) == 0)
    {
      if (!(f = PyDict_GetItemString(sysmodd, "stdin")))
	{
	  PyErr_SetString(PyExc_AttributeError, "sys.stdin isn't defined!");
	  return 0;
	}

      if (!PyFile_Check(f))
	{
	  PyErr_SetString(PyExc_TypeError, "sys.stdin isn't a file!");
	  return 0;
	}
    }
  else if (!PyArg_ParseTuple(args, "O!;sexpr.read", &PyFile_Type, &f))
    return 0;

  /* protect the ref since we allow threads during reads */
  Py_INCREF(f);

  o = read0(PyFile_AsFile(f));

  Py_DECREF(f);

  return o;
}

static PyObject *
getattr(PyObject *self, PyObject *args)
{
  char *s;

  fprintf(stderr, "in sexpr.__getattr__(%p)\n", args);

  if (!PyArg_ParseTuple(args, "s;sexpr.__getattr__"), &s)
    return 0;
  else if (strcmp(s, "null") == 0)
    {
      Py_INCREF(nullsexpr);
      return nullsexpr;
    }
  else
    return 0;
}

static PyMethodDef SexprMethods[] =
{
  {"sexpr", newsexpr, METH_VARARGS},
  {"sexprp", sexprp, METH_VARARGS},
  {"nullp", nullp, METH_VARARGS},
  {"pairp", pairp, METH_VARARGS},
  {"__getattr__", getattr, METH_VARARGS},
  {"cons", cons, METH_VARARGS},
  {"car", car, METH_VARARGS},
  {"cdr", cdr, METH_VARARGS},
  {"symbolp", symbolp, METH_VARARGS},
  {"eofobjectp", eofobjectp, METH_VARARGS},
  {"intern", intern, METH_VARARGS},
  {"read", readsexpr, METH_VARARGS},
  {"write", writesexpr, METH_VARARGS},
  {"display", writesexpr, METH_VARARGS},
  {0, 0, 0},
};

void
initsexpr()
{
  PyObject *m, *d;

  if (!mknullsexpr())
    Py_FatalError("can't initialize sexpr");

  if (!(eofobject = mktoksexpr()))
    Py_FatalError("can't initialize sexpr");
  if (!(openp = mktoksexpr()))
    Py_FatalError("can't initialize sexpr");
  if (!(closep = mktoksexpr()))
    Py_FatalError("can't initialize sexpr");
  if (!(dot = mktoksexpr()))
    Py_FatalError("can't initialize sexpr");

  if (!(obarray = PyDict_New()))
    Py_FatalError("can't initialize sexpr");

  if (!(sysmod = PyImport_ImportModule("sys")))
    Py_FatalError("can't initialize sexpr");
  if (!(sysmodd = PyModule_GetDict(sysmod)))
    Py_FatalError("can't initialize sexpr");

  m = Py_InitModule("sexpr", SexprMethods);
  d = PyModule_GetDict(m);
  SexprError = PyErr_NewException("sexpr.error", NULL, NULL);
  PyDict_SetItemString(d, "error", SexprError);
  if (PyErr_Occurred())
    Py_FatalError("can't initialize sexpr");
}
